(ns
  ^{:author "Brian Craft"
    :doc "Compact and interning string collections."}
  cavm.string-coll
  (:require [cavm.hashable :refer [ahashable get-array]]))

(defn str->arr [^String s]
  (ahashable (.getBytes s)))

(defn arr->str [hashable]
  (String. ^bytes (get-array hashable)))

(declare ->StringVec)

(deftype StringVec [pv cache]
  clojure.lang.IPersistentVector
  (seq [self] (map arr->str pv))
  (nth [self i] (arr->str (.nth ^clojure.lang.IPersistentVector pv i)))
  (nth [self i notfound] (arr->str (.nth ^clojure.lang.IPersistentVector pv i (str->arr notfound))))
  clojure.lang.ILookup
  (valAt [self i] (when-let [res (.valAt ^clojure.lang.IPersistentVector pv i)]
                    (arr->str res)))
  (valAt [self i notfound] (arr->str (.valAt ^clojure.lang.IPersistentVector pv i (str->arr notfound))))
  clojure.lang.ISeq
  (first [self] (arr->str (first pv)))
  (next [self] (->StringVec (next pv) cache))
  (more [self] (->StringVec (rest pv) cache))
  (cons [self s] (let [arr (str->arr s)]
                   (if-let [carr (cache arr)]
                     (->StringVec (.cons ^clojure.lang.IPersistentVector pv carr) cache) 
                     (->StringVec (.cons ^clojure.lang.IPersistentVector pv arr) (conj cache arr)))))
  (count [self] (count pv))
  Object
  (toString [self] (str (into [] self))))

(defn stringvec [coll]
  (into (->StringVec [] #{}) coll))

(defmethod print-method StringVec [v, ^java.io.Writer w]
  (.write w (.toString ^StringVec v)))
